require 'json'
require 'uu/os/lang/json_patch'
require 'uu/os/lang/uniform_hash'
require 'uu/os/lang/hash_accessor'
require 'uu/os/util/name_converter'
require 'uu/os/util/url_converter'
require 'uu/os/server/parameter_parser'
require 'uu/os/server/abstract_servlet'
require 'uu/os/ipc/ipc_error'
require 'uu/os/ipc/error_kind'
require 'uu/os/ipc/error_codes'
require 'uu/os/application/application_error'
require 'uu/os/application/error_codes'
require 'uu/os/application/controller_loader'
require 'uu/os/application/immediate_response'

module UU
  module OS
    module Application

      # Ancestor for all controllers which are agnostic to particular server
      # (in terms of technology) on which is controller deployed.
      class AbstractController < UU::OS::Server::AbstractServlet
        include UU::OS::Util::NameConverter
        include UU::OS::Server::ParameterParser

        # Main method for processing requests. By default it:
        #
        # * Asserts invocation method using {#assert_invocation_method} method.
        # * Asserts request content type using {#assert_content_type} method.
        # * Obtains main object using {#get_main_object_uri} method.
        # * Obtains parameters using {#get_parameters} method.
        # * Determine action method using {#get_action_method} method.
        # * Does necessary initializations using {#prepare_execution} method.
        # * Prepares context using {#get_context} method.
        # * Invokes action method.
        # * Converts action method result into expected response object.
        # * Does cleanup using {#finalize_execution} method.
        #
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [Object] Response object (same value as returned by {#response},
        #    {#file_download_response} or {#streamed_response} methods).
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def service(request) # :nodoc:
          begin
            assert_invocation_method(request)
            assert_content_type(request)
            uu_uri = get_main_object_uri(request)
            parameters = get_parameters(request)
            action_method = get_action_method(request, uu_uri, parameters)
            prepare_execution(request, uu_uri, parameters)
            context = get_context(request, uu_uri, parameters)
            result = self.send(action_method, context)
          rescue UU::OS::Application::ImmediateResponse => e
            return e.response
          rescue Exception => e
            # Log "raw" exception.
            log_error(context, e)
            # Allow inherited controllers to handle error.
            # (transform to another exception)
            error = handle_error(e) || e
            error_declared = false
            (declared_errors || []).each do |error_class|
              if error.kind_of?(error_class)
                error_declared = true
                break;
              end
            end
            if error_declared
              # Declared error should be passed directly to client.
              status = (error.respond_to?(:kind) && error.kind == UU::OS::IPC::ErrorKind::CLIENT) ? 400 : 500
              return response(status, nil, error.to_json)
            else
              # Wrap undeclared error to common error.
              if UU::OS::Server::ServerContext.development_mode?
                common_error = UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::SERVER, UU::OS::Application::ErrorCodes::UNEXPECTED_PROBLEM,
                	'Unexpected server error', error)
              else
                # Do not publish internal problem when not in development mode.
                common_error = UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::SERVER, UU::OS::Application::ErrorCodes::UNEXPECTED_PROBLEM,
                	'Unexpected server error')
              end
              # Use same exception ID to be able to find exception in logs.
              common_error.instance_variable_set(:@id, error.id)
              return response(500, nil, common_error.to_json)
            end
          ensure
            begin
              finalize_execution(request, uu_uri, parameters)
            rescue UU::OS::Application::ImmediateResponse => ef
              return ef.response
            rescue Exception => ef
              log_error(context, 'Finalization of request failed.', ef)
            end
          end
          if result.nil?
            return response(204, {}, result)
          elsif result.kind_of?(Hash) && result.has_key?(:status) && result.has_key?(:headers) && result.has_key?(:body)
            return result
          elsif result.respond_to?(:read)
            return response(200, {content_type: 'application/json'}, result)
          else
            return response(200, {content_type: 'application/json'}, result.to_json)
          end
        end

        # Checks if invocation method (POST, GET, etc.) matches controller type.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [void]
        # @raise [UU::OS::IPC::IPCError] If invocation method does not match.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def assert_invocation_method(request) # :nodoc:
          # No assertion for abstract controller
        end

        # Checks if incoming request body has valid content type.
        # By default +application/json+ or +multipart/form-data+
        # is expected.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [void]
        # @raise [UU::OS::IPC::IPCError] In case content type does not match
        #    expected one.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def assert_content_type(request) # :nodoc:
          if request.method == :post && !request.headers.contains(:content_type, '(application\/json)|(multipart\/form-data)')
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::INVALID_REQUEST,
            	"Invalid content type, expected application/json or multipart/form-data.")
          end
        end

        # Checks if binary value upload is expected by this controller.
        # In case method returns true, +data+ attribute from incoming
        # multipart request will be handed as direct socket stream. Data
        # attribute is also expected to be last item of multipart message
        # (any following items will be ignored).
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [TrueClass, FalseClass] True if binary upload is expected.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def binary_expected?(request) # :nodoc:
          return false
        end

        # Returns name of main object URI parameter (defaults to +:uuUri+).
        # @return [Symbol] Main object URI parameter name.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def main_object_uri_parameter_name() # :nodoc:
          return :uuUri
        end

        # Returns name of parameter which is used for wrapping of parameters
        # sent in get requests (defaults to +:uuDto+).
        # @return [Symbol] Parameters parameter name.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def parameters_parameter_name() # :nodoc:
          return :uuDto
        end

        # Returns name of parameter which is used for wrapping of invocation
        # context parameters (defaults to +:uuCtx+).
        # @return [Symbol] Context parameter name.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def context_parameter_name() # :nodoc:
          return :uuCtx
        end

        # Returns main object URI from request. By default main object URI is obtained
        # from query parameters using key returned by {#main_object_uri_parameter_name}.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [String] Main object URI.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def get_main_object_uri(request) # :nodoc:
          return request.parameters[main_object_uri_parameter_name]
        end

        # Parses parameters from request. By default query parameters (with exclusion
        # of {#main_object_uri_parameter_name} parameter) are merged with parameters
        # sent in body. In case query parameter or form parameter (from multipart request)
        # contains JSON value, it is also transparently deserialized.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [Hash] Parsed and deserialized parameters.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def get_parameters(request) # :nodoc:
          begin
            parameters = parse_parameters(request, (binary_expected?(request) ? :data : nil))
          rescue TypeError => e
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::INVALID_REQUEST,
            	"Invalid request: #{e.message}")
          end
          # Optionally unwrap get parameters
          if request.method == :get && !parameters_parameter_name.nil? && !parameters_parameter_name.empty?
            wrapped = parameters.delete(parameters_parameter_name)
            if !wrapped.nil? && !wrapped.kind_of?(Hash)
              raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::INVALID_REQUEST,
              	"Invalid request: parameter #{parameters_parameter_name} must contain key-value structure")
            else
              parameters.merge!(wrapped) if !wrapped.nil?
            end
          end
          # Remove main object URI parameter
          parameters.delete(main_object_uri_parameter_name)
          return parameters
        end

        # Returns name of action method to be invoked on controller.
        # by default action method is constructed from last request
        # path element converted to +snake_case+ with +on_+ prefix.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @param [String] uu_uri URI of main object.
        # @param [Hash] parameters Parsed and deserialized parameters.
        # @return [Symbol, String] Name of action method to be invoked.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def get_action_method(request, uu_uri, parameters) # :nodoc:
          action = request.url.split('/')[-1]
          action_method = 'on_' << to_snake_case(action)
          return action_method.to_sym
        end

        # Prepares context object.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @param [String] uu_uri URI of main object.
        # @param [Hash] parameters parsed parameters.
        # @return [Object] Execution context
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def get_context(request, uu_uri, parameters) # :nodoc:
          # Prepare mock execution context object
          context_class = Class.new(UU::OS::Lang::UniformHash)
          context_class.send(:include, UU::OS::Lang::HashAccessor)
          context_class.send(:include, UU::OS::Util::NameConverter)
          class << context_class
            protected
            def transform_key(key)
              return nil if key.nil?
              return to_snake_case(key).to_sym
            end
          end
          context = context_class.new
          context[:request] = request
          context[:uu_uri] = uu_uri
          context[:parameters] = parameters
          context[:'_$ctx'] = parameters.delete(context_parameter_name)
          return context
        end

        # Method is invoked before execution of control method. It should
        # be used for necessary initializations of environment. In case
        # method is overridden, new implementation should always use +super+
        # call to ensure all initializations are done.
        # @param [UU::OS::CON::HTTPRequest] request Incoming request.
        # @param [String] uu_uri URI of main object.
        # @param [Hash] parameters parsed parameters.
        # @return [void]
        # @raise [Exception] In case initialization of environment fails.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def prepare_execution(request, uu_uri, parameters) # :nodoc:

        end

        # Method is always invoked after execution of control method (even if
        # method raised exception). It should be used to cleanup environment
        # (cleanup everything what was initialized in {#prepare_execution}
        # method). In case method is overridden, new implementation should always
        # use +super+ call to ensure cleanup is complete. This method must not
        # raise any exception.
        # @param [UU::OS::CON::HTTPRequest] request Incoming request.
        # @param [String, nil] uu_uri URI of main object or +nil+ if
        #    method is invoked due to error before request parameters are parsed.
        # @param [Hash, nil] parameters Parsed parameters or +nil+ if method is
        #    invoked before request parameters are parsed.
        # @return [void]
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def finalize_execution(request, uu_uri, parameters) # :nodoc:

        end

        # Method returns list of errors which are declared for this controller
        # (are defined as part of API) and thus are directly passed to client
        # side. In case controller raises any other exception, it is wrapped as
        # unhandled IPC error.
        # @return [Array<Class<UU::OS::IPC::IPCError>>] List of declared errors.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def declared_errors # :nodoc:
          return [UU::OS::IPC::IPCError]
        end

        protected

        # Method is invoked when exception is caught. It allows to handle this
        # error by transforming is to another error. It is recommended to reuse
        # exception ID to be able to easily trace original error in logs.
        # @param [Exception] Caught exception.
        # @return [Exception] Exception to be raised.
        # @private
        def handle_error(exception) # :nodoc:

        end

        # Method is invoked whenever unhandled exception is raised from
        # controller implementation. This method should be overridden
        # in particular controller based on available technology. By
        # default error is logged to STDERR.
        # @private
        def log_error(context, message = nil, exception = nil) # :nodoc:
          message, exception = exception, message if message.kind_of?(Exception)
          return if message.nil? && exception.nil?
          msg = "ERROR #{Time.now} "
          if !message.nil? && !message.empty?
            msg << message.to_s
          end
          msg << "\n  " if !exception.nil?
          if !exception.nil?
            msg << "#{exception.class} [#{exception.id}]: #{exception.message}"
            trace = exception.backtrace
            msg << "\n  " << trace.join("\n  ") if !trace.nil? && !trace.empty?
          end
          STDERR.puts(msg)
        end

        private

        # Transforms invocation of unimplemented controller action method
        # into valid exception response.
        # @private
        def method_missing(method, *args, &block) # :nodoc:
          if method =~ /^on_/ && args.size == 1
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::NOT_IMPLEMENTED,
            	"Functionality addressed by #{args[0].request.url} is not implemented.")
          else
            super(method, *args)
          end
        end

      end
    end
  end
end
